"""Provides utilities for scanning the music library filesystem for files.

This module provides a LibraryScanner class that can be used to walk through a directory and
pick up any files relevant to musicbird, plus changes made to it.
"""

import logging
from pathlib import Path

import ffmpeg

from .db import LibraryDB
from .file import File


logger = logging.getLogger(__name__)


class LibraryScanner:
    """Index a directory and add its contents to the DB.

    Use this class to scan a directory for files and add them to the provided Database.
    """

    def __init__(self, path: Path, db: LibraryDB) -> None:
        """Generate a new scanner for the given path and database.

        Args:
            path (Path): The path to scan.
            db (LibraryDB): The library to save the result to.
        """
        self.path = path.resolve()
        self.db = db

    def scan(self) -> bool:
        """ Scan the filesystem for changes.

        Registers new, modified and deleted files and updates the Library.

        Returns:
            bool: True if all files were scanned successfully, False if not.
        """
        successful = False

        # Mark all files as deleted in-memory at the beginning of our scan
        # As files are rediscovered, the deleted flag will be removed
        logger.info(f"Scanning directory {self.path} into library")

        for file in self.db.get_all_files():
            file.was_deleted = True
            self.db.add_or_update_file(file)

        # todo: replace glob with glob from pathlib. Haven't found a way to make that work yet
        for path in self.path.glob("./**/*"):
            if path.is_file():
                file = File(path)
                try:
                    file.determine_type()
                except ffmpeg.Error as e:
                    logger.error(f"Could not determine filetype, skipping: {file.path}. FFmpeg error: {e.stderr}")
                    successful = False
                else:
                    self.add_or_update_file(file)
        successful = True
        return successful

    def add_or_update_file(self, file: File) -> None:
        """Adds a single new file to the library or update an existing one.

        Args:
            file (File): The file object to add
        """
        # If we're adding/updating a file, that means it must also exist physically
        file.was_deleted = False

        current_entry = self.db.get_file_by_path(file.path)
        if not current_entry:
            logger.info(f"Adding new file to library: {file.path}")
            file.needs_processing = True
        elif current_entry.mtime != file.mtime or current_entry.type != file.type:
            logger.info(f"Existing file has been modified and will be reprocessed: {file.path}")
            file.needs_processing = True
        else:
            logger.debug(f"File unchanged since last scan: {file.path}")
        self.db.add_or_update_file(file)



"""Interfaces for encoding audio files.

This module provides the interface used by the other MusicBird components to encode audio files.
The Abstract class Encoder serves as the interface used by other components, while individual Encoder implementations
(such as MP3Encoder) are derived from it.
"""
from abc import ABC, abstractmethod
import logging
from pathlib import Path
import subprocess
import sys
from typing import Dict

import ffmpeg

logger = logging.getLogger(__name__)


class Encoder(ABC):
    """Interface for interacting with an Audio file Encoder.

    All encode operations are intended to be done using the methods defined by this class
    """

    extension = ""

    @abstractmethod
    def encode(self, src: Path, dest: Path) -> bool:
        """Encode the file at src to dest.

        Calls the encoders backend and converts the source file, then saves the output at the destination directory.

        Args:
            src (Path): The file to encode.
            dest (Path): The Path at which to store the encoded file.

        Returns:
            bool: True if the operation was successful, False if not.
        """


class FFmpegEncoder(Encoder):
    """Base class for all encoders utilizing ffmpeg.

    Child encoders can inherit from this class and set their parameters accordingly.
    """
    # We might need to use config in this class in the future, so keep it in
    # pylint: disable=unused-argument

    def __init__(self, config: Dict) -> None:
        super().__init__()
        self.ffmpeg_args = {}

        # Check if we can access ffmpeg
        try:
            subprocess.run(["ffmpeg", "-version"], stdout=subprocess.PIPE, stderr=subprocess.PIPE, check=True)
        except FileNotFoundError as e:
            logger.fatal("Could not access ffmpeg. Pease make sure that it is installed")
            raise e
        except subprocess.CalledProcessError as e:
            logger.fatal(f"Could not verify that ffmpeg is ready. Error: {repr(e)}")
            raise e

    def encode(self, src: Path, dest: Path) -> bool:
        dest = dest.with_suffix(self.extension)
        try:
            dest.parent.mkdir(parents=True, exist_ok=True)
        except OSError as e:
            logger.error(f"Could not create directory to encode file {dest}: {repr(e)}")
            return False
        try:
            stream = ffmpeg.input(str(src))
            stream = ffmpeg.output(stream, str(dest), **self.ffmpeg_args)
            stream = ffmpeg.overwrite_output(stream)
            ffmpeg.run(stream, quiet=True)
        except ffmpeg.Error as e:
            logger.error(f"Failed to encode file {src}. ffmpeg error: \n {e.stderr}")
            return False
        return True


class MP3Encoder(FFmpegEncoder):
    def __init__(self, config: Dict) -> None:
        super().__init__(config)
        self.extension = ".mp3"
        self.ffmpeg_args["acodec"] = "libmp3lame"
        if config["vbr"]:
            self.ffmpeg_args["q:a"] = config["quality"]
        else:
            self.ffmpeg_args["audio_bitrate"] = config["bitrate"]


class OpusEncoder(FFmpegEncoder):
    def __init__(self, config: Dict) -> None:
        super().__init__(config)
        self.extension = ".opus"
        self.ffmpeg_args["acodec"] = "libopus"
        self.ffmpeg_args["audio_bitrate"] = config["bitrate"]


def init(config, exit_on_error: bool = True) -> Encoder:
    """Initialize an encoder object based on the values provided in the config.

    Creates an apropiate LibraryDB object for the given backend in config and returns it.

    Args:
        config (Dict): MusicBirds configuration
        exit_on_error (bool, optional): Whether to call sys.exit() if the initialization fails.
            If set to false, will raise exception instead. Defaults to False.

    Returns:
        Encoder: The Encoder object
    """
    try:
        if config["encoder"] == "mp3":
            return MP3Encoder(config["mp3"])
        elif config["encoder"] == "opus":
            return OpusEncoder(config["opus"])
    except (OSError, ffmpeg.Error) as e:
        if exit_on_error:
            sys.exit("Error initializing database")
        else:
            raise e
